import os
from typing import IO, Any, Dict, List, Tuple

import numpy as np
from tqdm import tqdm

from . import ioutils as io
from .discretization import Discretization

shape_to_eletype: Dict[str, str] = {
    "TET4": "tetra4",
    "TET10": "tetra10",
    "HEX8": "hexa8",
    "HEX20": "hexa20",
    "HEX27": "hexa27",
    "QUAD4": "quad4",
    "TRI3": "tria3",
    "LINE2": "bar2",
}


def get_unique_filename(filename: str) -> str:
    splitted = os.path.splitext(filename)

    # write geometry file
    fname = "{0}{1}".format(splitted[0], splitted[1])
    i = 1
    while os.path.exists(fname):
        fname = "{0}-{1}{2}".format(splitted[0], i, splitted[1])
        i += 1

    return fname


def write_case(
    filename,
    dis: Discretization,
    binary: bool = True,
    override: bool = False,
    out: bool = True,
) -> None:
    dis.compute_ids(False)
    filedir = os.path.dirname(filename)

    basename = os.path.splitext(os.path.basename(filedir))[0]

    # write geometry file
    if override:
        geofile = os.path.join(filedir, "{0}_geometry.geo".format(basename))
    else:
        geofile = get_unique_filename(
            os.path.join(filedir, "{0}_geometry.geo".format(basename))
        )

    with open(geofile, "w{0}".format("b" if binary else "")) as f:
        write_geometry(f, dis, binary=binary, out=out)

    # A variable is supposed to be transient if len(shape) == 3 (time component is first index)
    # write element variables
    ele_count = {}
    # get number of elements for each type
    for eles in dis.elements.values():
        for ele in eles:
            if ele.shape not in shape_to_eletype:
                raise NotImplementedError(
                    "This kind of element is not known: {0}".format(ele.shape)
                )
            eletype = shape_to_eletype[ele.shape]

            if eletype not in ele_count:
                ele_count[eletype] = 0

            ele_count[eletype] += 1

    # build element variables
    ele_vars = {}
    ele_cur_count = {}
    for eles in tqdm(
        dis.elements.values(), disable=not out, desc="Prepare element data"
    ):
        for ele in eles:
            if ele.shape not in shape_to_eletype:
                raise NotImplementedError(
                    "This kind of element is not known: {0}".format(ele.shape)
                )
            eletype = shape_to_eletype[ele.shape]

            if eletype not in ele_cur_count:
                ele_cur_count[eletype] = 0

            for varname, data in ele.data.items():
                data = np.array(data)
                if len(data.shape) == 0:
                    data = data.reshape((1))

                if varname not in ele_vars:
                    ele_vars[varname] = {}
                    for elt, eltc in ele_count.items():
                        ele_vars[varname][elt] = np.zeros(
                            tuple([eltc] + list(data.shape)), dtype=data.dtype
                        )

                ele_vars[varname][eletype][ele_cur_count[eletype]] = data

            ele_cur_count[eletype] += 1

    # write them finally
    ele_vars_props = {}
    for varname, data in tqdm(
        ele_vars.items(), disable=not out, desc="Write element data"
    ):
        if override:
            varfile = os.path.join(
                filedir, "{0}_variable.{1}".format(basename, varname)
            )
        else:
            varfile = get_unique_filename(
                os.path.join(filedir, "{0}_variable.{1}".format(basename, varname))
            )
        ele_vars_props[varname] = {}
        ele_vars_props[varname]["filename"] = os.path.basename(varfile)
        with open(varfile, "w{0}".format("b" if binary else "")) as f:
            (
                ele_vars_props[varname]["timesteps"],
                ele_vars_props[varname]["dim"],
            ) = write_element_variable(f, dis, varname, data, binary=binary)

    # write nodal variables
    nodal_vars = {}
    for i, node in tqdm(
        enumerate(dis.nodes), disable=not out, desc="Prepare nodal data"
    ):
        for varname, data in node.data.items():
            # ensure that data is a np array
            data = np.array(data)
            if len(data.shape) == 0:
                data = data.reshape((1))
            if varname not in nodal_vars:
                nodal_vars[varname] = np.zeros(
                    tuple([len(dis.nodes)] + list(data.shape)),
                    dtype=data.dtype,
                )

            nodal_vars[varname][i] = data

    nodal_vars_props = {}
    for varname, data in tqdm(
        nodal_vars.items(), disable=not out, desc="Write nodal data"
    ):
        if override:
            varfile = os.path.join(
                filedir, "{0}_variable.{1}".format(basename, varname)
            )
        else:
            varfile = get_unique_filename(
                os.path.join(filedir, "{0}_variable.{1}".format(basename, varname))
            )
        nodal_vars_props[varname] = {}
        nodal_vars_props[varname]["filename"] = os.path.basename(varfile)
        with open(varfile, "w{0}".format("b" if binary else "")) as f:
            (
                nodal_vars_props[varname]["timesteps"],
                nodal_vars_props[varname]["dim"],
            ) = write_node_variable(f, dis, varname, data, binary=binary)

    # write case file
    if override:
        casefile = filename
    else:
        casefile = get_unique_filename(filename)
    with open(casefile, "w") as f:
        _write_case(f, os.path.basename(geofile), ele_vars_props, nodal_vars_props)


def _write_case(
    fstream: IO,
    geofile: str,
    ele_vars_props: Dict[str, Dict[str, Any]],
    nodal_vars_props: Dict[str, Dict[str, Any]],
) -> None:
    sets = {1: 1}
    dim_to_type = {1: "scalar", 3: "vector", 6: "tensor symm"}

    nextset = 2
    for prop in ele_vars_props.values():
        if prop["timesteps"] not in sets:
            sets[prop["timesteps"]] = nextset
            nextset += 1

    for prop in nodal_vars_props.values():
        if prop["timesteps"] not in sets:
            sets[prop["timesteps"]] = nextset
            nextset += 1

    fstream.write("FORMAT\n\ntype:\tensight gold\n\n")
    fstream.write("GEOMETRY\n\n")
    fstream.write("model:\t{0}\t{1}\t{2}\n\n\n".format(sets[1], sets[1], geofile))
    fstream.write("VARIABLE\n\n")
    # Todo variables
    for varname, props in ele_vars_props.items():
        if props["dim"] not in dim_to_type:
            continue
        fstream.write(
            "{0} per {1}:\t{2}\t{3}\t{4}\t{5}\n".format(
                dim_to_type[props["dim"]],
                "element",
                sets[props["timesteps"]],
                sets[props["timesteps"]],
                varname,
                props["filename"],
            )
        )

    for varname, props in nodal_vars_props.items():
        if props["dim"] not in dim_to_type:
            continue
        fstream.write(
            "{0} per {1}:\t{2}\t{3}\t{4}\t{5}\n".format(
                dim_to_type[props["dim"]],
                "node",
                sets[props["timesteps"]],
                sets[props["timesteps"]],
                varname,
                props["filename"],
            )
        )

    fstream.write("\n")
    fstream.write("TIME\n\n")
    for id, count in sets.items():
        fstream.write(get_timeset_entry(id, count))
        fstream.write("\n")

    fstream.write("\n")
    fstream.write("FILE\n\n")
    for id, count in sets.items():
        fstream.write("file set:\t\t{0}\nnumber of steps:\t\t{1}\n".format(id, count))
    fstream.write("\n")


def get_timeset_entry(id, count):
    entry: str = "time set:\t\t{0}\nnumber of steps:\t\t{1}\ntime values: ".format(
        id, count
    )

    i: int = 0
    for item in range(count):
        if i > 0 and i % 8 == 0:
            entry += "\n"

        entry += " {0}".format(float(item))

        i += 1

    return entry


def write_geometry(
    fstream: IO, dis: Discretization, binary: bool = True, out: bool = True
) -> None:
    if binary:
        io.ens_write_string(fstream, "C Binary", binary=True)

    io.ens_write_string(fstream, "BEGIN TIME STEP", binary=binary)
    io.ens_write_string(fstream, "geometry", binary=binary)
    io.ens_write_string(fstream, "Comment", binary=binary)
    io.ens_write_string(fstream, "node id given", binary=binary)
    io.ens_write_string(fstream, "element id off", binary=binary)  # should I switch on?
    io.ens_write_string(fstream, "part", binary=binary)
    io.ens_write_int(fstream, 1, binary=binary)
    io.ens_write_string(fstream, "field", binary=binary)
    io.ens_write_string(fstream, "coordinates", binary=binary)
    io.ens_write_int(fstream, len(dis.nodes), binary=binary)
    io.ens_write_ints(fstream, np.array([n.id for n in dis.nodes]), binary=binary)
    io.ens_write_floats(fstream, dis.get_node_coords(), binary=binary)

    # build eletype array
    elegroups = {}
    for eles in tqdm(
        dis.elements.values(), disable=not out, desc="Prepare element geometry"
    ):
        for ele in eles:
            if ele.shape not in shape_to_eletype:
                raise NotImplementedError(
                    "This kind of element is not known: {0}".format(ele.shape)
                )
            eletype = shape_to_eletype[ele.shape]

            if eletype not in elegroups:
                # np.zeros((0,len(ele.nodes)), dtype=int)
                elegroups[eletype] = []

            elegroups[eletype].append(np.array([n.id for n in ele.nodes]))

    for eletype, eles in elegroups.items():
        io.ens_write_string(fstream, eletype, binary=binary)
        io.ens_write_int(fstream, len(eles), binary=binary)
        io.ens_write_ints(fstream, np.array(eles, dtype=int), binary=binary)

    io.ens_write_string(fstream, "END TIME STEP", binary=binary)


def write_element_variable(
    fstream: IO,
    dis: Discretization,
    varname: str,
    data: Dict[str, np.ndarray],
    binary: bool = True,
) -> Tuple[int, int]:
    if binary:
        io.ens_write_string(fstream, "C Binary", binary=True)

    dim = None
    ts = None

    for key, edata in data.items():
        if len(edata.shape) > 3:
            raise RuntimeError("This kind of data is not supported. To many dimensions")

        if len(edata.shape) == 3:
            # This will be interpreted as transient vector data
            data[key] = np.transpose(edata, (1, 0, 2))

        if len(edata.shape) == 2:
            # This will be interpreted as non transient vector data
            data[key] = edata.reshape(tuple([1] + list(edata.shape)))

        if len(edata.shape) == 1:
            # This will be interpreted as non transient scalar data
            data[key] = edata.reshape(tuple([1] + list(edata.shape) + [1]))

        # safety check
        if dim is None:
            dim = data[key].shape[2]
            ts = data[key].shape[0]
        elif dim != data[key].shape[2]:
            raise RuntimeError(
                "The elemnt data of variable {0} have not the correct dimesion accross all elements".format(
                    varname
                )
            )
        elif ts != data[key].shape[0]:
            raise RuntimeError(
                "The elemnt data of variable {0} have not the correct num_timesteps accross all elements".format(
                    varname
                )
            )

    if ts is None:
        raise RuntimeError("Timestep is None")
    if dim is None:
        raise RuntimeError("dim is None")

    for t in range(ts):
        io.ens_write_string(fstream, "BEGIN TIME STEP", binary=binary)
        io.ens_write_string(fstream, "description", binary=binary)
        io.ens_write_string(fstream, "part", binary=binary)
        io.ens_write_int(fstream, 1, binary=binary)

        for k, d in data.items():
            io.ens_write_string(fstream, k, binary=binary)
            io.ens_write_floats(fstream, d[t], binary=binary)

        io.ens_write_string(fstream, "END TIME STEP", binary=binary)

    return ts, dim


def write_node_variable(
    fstream: IO,
    dis: Discretization,
    varname: str,
    data: np.ndarray,
    binary: bool = True,
) -> Tuple[int, int]:
    if binary:
        io.ens_write_string(fstream, "C Binary", binary=True)

    if len(data.shape) > 3:
        raise RuntimeError("This kind of data is not supported. To many dimensions")

    if len(data.shape) == 3:
        # This will be interpreted as transient vector data
        data = np.transpose(data, (1, 0, 2))

    if len(data.shape) == 2:
        # This will be interpreted as non transient vector data
        data = data.reshape(tuple([1] + list(data.shape)))

    if len(data.shape) == 1:
        # This will be interpreted as non transient scalar data
        data = data.reshape(tuple([1] + list(data.shape) + [1]))

    # safety check
    if data.shape[1] != len(dis.nodes):
        raise RuntimeError(
            "The nodal data of variable {0} does not fit".format(varname)
        )

    for tdata in data:
        io.ens_write_string(fstream, "BEGIN TIME STEP", binary=binary)
        io.ens_write_string(fstream, "description", binary=binary)
        io.ens_write_string(fstream, "part", binary=binary)
        io.ens_write_int(fstream, 1, binary=binary)
        io.ens_write_string(fstream, "coordinates", binary=binary)
        io.ens_write_floats(fstream, tdata, binary=binary)
        io.ens_write_string(fstream, "END TIME STEP", binary=binary)

    return data.shape[0], data.shape[2]



import os
import unittest

import lnmmeshio
import numpy as np
from lnmmeshio import Fiber

script_dir = os.path.dirname(os.path.realpath(__file__))


class TestFibers(unittest.TestCase):
    def setUp(self):
        pass

    def test_tri3_integration(self):
        tri3 = lnmmeshio.Tri3(
            "T3",
            [
                lnmmeshio.Node(np.array([0.0, 0.0, 0.0])),
                lnmmeshio.Node(np.array([1.0, 0.0, 0.0])),
                lnmmeshio.Node(np.array([2.0, 3.0, 0.0])),
            ],
        )

        np.testing.assert_almost_equal(
            tri3.integrate(lambda x: 1.1, 1), np.array(1.5 * 1.1)
        )
        np.testing.assert_almost_equal(
            tri3.integrate(lambda x: 1.1, 3), np.array(1.5 * 1.1)
        )
        np.testing.assert_almost_equal(
            tri3.integrate_xi(lambda xi: 1.1, 1), np.array(1.5 * 1.1)
        )
        np.testing.assert_almost_equal(
            tri3.integrate_xi(lambda xi: 1.1, 3), np.array(1.5 * 1.1)
        )

        tri3 = lnmmeshio.Tri3(
            "T3",
            [
                lnmmeshio.Node(np.array([0.0, 0.0, 1.0])),
                lnmmeshio.Node(np.array([1.0, 0.0, 4.0])),
                lnmmeshio.Node(np.array([2.0, 3.0, 5.0])),
            ],
        )

        np.testing.assert_almost_equal(
            tri3.integrate(lambda x: 1.1, 1), np.array(5.3324478)
        )
        np.testing.assert_almost_equal(
            tri3.integrate(lambda x: 1.1, 3), np.array(5.3324478)
        )

        np.testing.assert_almost_equal(
            tri3.integrate_xi(lambda xi: 1.1, 1), np.array(5.3324478)
        )
        np.testing.assert_almost_equal(
            tri3.integrate_xi(lambda xi: 1.1, 3), np.array(5.3324478)
        )

    def test_tri6_integration(self):
        tri6 = lnmmeshio.Tri6(
            "T6",
            [
                lnmmeshio.Node(np.array([0.0, 0.0, 0.0])),
                lnmmeshio.Node(np.array([1.0, 0.0, 0.0])),
                lnmmeshio.Node(np.array([2.0, 3.0, 0.0])),
                lnmmeshio.Node(np.array([0.5, 0.0, 0.0])),
                lnmmeshio.Node(np.array([1.5, 1.5, 0.0])),
                lnmmeshio.Node(np.array([1.0, 1.5, 0.0])),
            ],
        )

        np.testing.assert_almost_equal(
            tri6.integrate(lambda x: 1.1, 1), np.array(1.5 * 1.1)
        )
        np.testing.assert_almost_equal(
            tri6.integrate(lambda x: 1.1, 3), np.array(1.5 * 1.1)
        )
        np.testing.assert_almost_equal(
            tri6.integrate_xi(lambda xi: 1.1, 1), np.array(1.5 * 1.1)
        )
        np.testing.assert_almost_equal(
            tri6.integrate_xi(lambda xi: 1.1, 3), np.array(1.5 * 1.1)
        )

        tri6 = lnmmeshio.Tri6(
            "T6",
            [
                lnmmeshio.Node(np.array([0.0, 0.0, 1.0])),
                lnmmeshio.Node(np.array([1.0, 0.0, 4.0])),
                lnmmeshio.Node(np.array([2.0, 3.0, 5.0])),
                lnmmeshio.Node(np.array([0.5, 0.0, 2.5])),
                lnmmeshio.Node(np.array([1.5, 1.5, 4.5])),
                lnmmeshio.Node(np.array([1.0, 1.5, 3.0])),
            ],
        )

        np.testing.assert_almost_equal(
            tri6.integrate(lambda x: 1.1, 1), np.array(5.3324478)
        )
        np.testing.assert_almost_equal(
            tri6.integrate(lambda x: 1.1, 3), np.array(5.3324478)
        )
        np.testing.assert_almost_equal(
            tri6.integrate_xi(lambda xi: 1.1, 1), np.array(5.3324478)
        )
        np.testing.assert_almost_equal(
            tri6.integrate_xi(lambda xi: 1.1, 3), np.array(5.3324478)
        )

    def test_tri3_integration_x_xi(self):
        tri3 = lnmmeshio.Tri3(
            "T3",
            [
                lnmmeshio.Node(np.array([-1.0, -2.0, 0.0])),
                lnmmeshio.Node(np.array([2.0, -2.0, 0.0])),
                lnmmeshio.Node(np.array([-1.0, 2.0, 0.0])),
            ],
        )

        np.testing.assert_almost_equal(
            tri3.integrate(lambda x: 1.1 * x[0] + 1.2 * x[1], 1), np.array(-4.8)
        )
        np.testing.assert_almost_equal(
            tri3.integrate(lambda x: 1.1 * x[0] + 1.2 * x[1], 3), np.array(-4.8)
        )

        np.testing.assert_almost_equal(
            tri3.integrate_xi(
                lambda xi: 1.1 * (3 * xi[0] - 1) + 1.2 * (4 * xi[1] - 2), 1
            ),
            np.array(-4.8),
        )
        np.testing.assert_almost_equal(
            tri3.integrate_xi(
                lambda xi: 1.1 * (3 * xi[0] - 1) + 1.2 * (4 * xi[1] - 2), 3
            ),
            np.array(-4.8),
        )

    def test_tri6_integration_x_xi(self):
        tri6 = lnmmeshio.Tri6(
            "T6",
            [
                lnmmeshio.Node(np.array([-1.0, -2.0, 0.0])),
                lnmmeshio.Node(np.array([2.0, -2.0, 0.0])),
                lnmmeshio.Node(np.array([-1.0, 2.0, 0.0])),
                lnmmeshio.Node(np.array([0.5, -2.0, 0.0])),
                lnmmeshio.Node(np.array([0.5, 0.0, 0.0])),
                lnmmeshio.Node(np.array([-1.0, 0.0, 0.0])),
            ],
        )

        np.testing.assert_almost_equal(
            tri6.integrate(lambda x: 1.1 * x[0] + 1.2 * x[1], 1), np.array(-4.8)
        )
        np.testing.assert_almost_equal(
            tri6.integrate(lambda x: 1.1 * x[0] + 1.2 * x[1], 3), np.array(-4.8)
        )

        np.testing.assert_almost_equal(
            tri6.integrate_xi(
                lambda xi: 1.1 * (3 * xi[0] - 1) + 1.2 * (4 * xi[1] - 2), 1
            ),
            np.array(-4.8),
        )
        np.testing.assert_almost_equal(
            tri6.integrate_xi(
                lambda xi: 1.1 * (3 * xi[0] - 1) + 1.2 * (4 * xi[1] - 2), 3
            ),
            np.array(-4.8),
        )



import io
import math
from typing import Dict, List, Optional, Union

import numpy as np

from ..fiber import Fiber
from ..ioutils import (
    line_option_list,
    read_option_item,
    write_option,
    write_option_list,
    write_title,
)
from ..node import Node


class Element:
    """
    Class holding the data of one element
    """

    def __init__(
        self,
        el_type: Optional[str],
        shape: str,
        nodes: List[Node],
        options: Optional[dict] = None,
    ):
        """
        Creates a new element of type ele_type, shape with nodes defined in nodes

        Args:
            el_type: str of element type as used by BACI (e.g. SOHEX8)
            shape: shape of the element as used by BACI (e.g. HEX8)
            nodes: List of node objects
        """
        self.id: Optional[int] = None
        self.type = el_type
        self.shape = shape
        self.nodes = nodes
        self.options = options if options is not None else {}
        self.fibers: Dict[str, Fiber] = {}
        self.data: Dict[str, Union[np.ndarray, int, float]] = {}

    @classmethod
    def get_num_nodes(cls) -> int:
        """
        Get number of nodes of the element

        Returns:
            Number of nodes of the element
        """
        raise NotImplementedError("This method is not implemented in your element")

    def get_node_coords(self) -> np.ndarray:
        arr = np.zeros((self.get_num_nodes(), 3))

        for i, node in enumerate(self.nodes):
            arr[i, :] = node.coords

        return arr

    def reset(self):
        """
        Sets the element id to None
        """
        self.id = None

    def get_nodes(self) -> List[Node]:
        """
        Returns a list of nodes

        Returns:
            List of nodes
        """
        return self.nodes

    def get_faces(self) -> list:
        """
        Returns all faces

        Returns:
            List of faces
        """
        raise RuntimeError("This element does't implement get faces")

    def get_edges(self) -> list:
        """
        Returns all edges

        Returns:
            List of edges
        """
        raise RuntimeError("This element does't implement get edges")

    def get_node_ids(self):
        """
        Returns the node ids as of each element as a numpy array of shape (num_ele, num_nod_per_ele)

        Returns:
            np.array with node ids of each element
        """
        arr: np.ndarray = np.zeros((len(self.nodes)), dtype=int)

        for i, node in enumerate(self.nodes, start=0):
            if node.id is None:
                raise RuntimeError("You need to compute ids first")
            arr[i] = node.id

        return arr

    def get_dpoints(self):
        """
        Returns a list of dpoints that is shared by all nodes of the element

        Returns:
            List of dpoint
        """
        pointnodesets = None

        for p in self.nodes:
            if pointnodesets is None:
                pointnodesets = set(p.pointnodesets)
            else:
                pointnodesets = set(p.pointnodesets).intersection(pointnodesets)

        return list(pointnodesets)

    def get_dlines(self):
        """
        Returns a list of dlines that is shared by all nodes of the element

        Returns:
            List of dlines
        """
        linenodesets = None

        for p in self.nodes:
            if linenodesets is None:
                linenodesets = set(p.linenodesets)
            else:
                linenodesets = set(p.linenodesets).intersection(linenodesets)

        return list(linenodesets)

    def get_dsurfs(self):
        """
        Returns a list of dsurf that is shared by all nodes of the element

        Returns:
            List of dsurfs
        """
        surfacenodesets = None

        for p in self.nodes:
            if surfacenodesets is None:
                surfacenodesets = set(p.surfacenodesets)
            else:
                surfacenodesets = set(p.surfacenodesets).intersection(surfacenodesets)

        return list(surfacenodesets)

    def get_dvols(self):
        """
        Returns a list of dvol that is shared by all nodes of the element

        Returns:
            List of dvols
        """
        volumenodesets = None

        for p in self.nodes:
            if volumenodesets is None:
                volumenodesets = set(p.volumenodesets)
            else:
                volumenodesets = set(p.volumenodesets).intersection(volumenodesets)

        return list(volumenodesets)

    def get_line(self):
        line = io.StringIO()
        if self.id is None:
            raise RuntimeError("You have to compute ids before writing")

        line.write("{0} {1} ".format(self.id, self.type))

        options: dict = {}
        options[self.shape] = [i.id for i in self.nodes]
        options.update(self.options)

        line.write(line_option_list(options))

        for t, f in self.fibers.items():
            line.write(" ")
            line.write(f.get_line(t))

        return line.getvalue()

    def write(self, dest):
        """
        Write the corresponding element line in the dat file
        """
        if self.id is None:
            raise RuntimeError("You have to compute ids before writing")

        dest.write("{0} {1} ".format(self.id, self.type))

        options: dict = {}
        options[self.shape] = [i.id for i in self.nodes]
        options.update(self.options)

        write_option_list(dest, options, newline=False)

        for t, f in self.fibers.items():
            dest.write(" ")
            f.write(dest, t)

        dest.write("\n")

    @classmethod
    def get_space_dim(cls):
        """
        Returns the number of space dimensions

        Returns:
            Number of space dimensions
        """
        raise NotImplementedError(
            "This element is not correctly implemented. It needs to be derived from Element{1,2,3}D"
        )

    def get_xi(self, x):
        """
        Returns the local variables from given global variables

        Args:
            x: Global variable

        Returns:
            Local variables (xi)
        """
        raise NotImplementedError("This is currently not implemented for all elements")

    @classmethod
    def is_in_ref(cls, xi, include_boundary=True):
        """
        Checkes whether a point in reference coordinates is within the element

        Args:
            xi: Point in reference coordinates

        Returns:
            True if the point is within the element (or on the bounrary), otherwise False
        """
        raise NotImplementedError("This is currently not implemented for all elements")

    def is_in(self, x, include_boundary=True):
        """
        Check whether a point is within the element

        Args:
            x: Point in space
            include_boundary: bool, Whether or not to include the boundary

        Returns:
            True if the point is within the element (or on the boundary), otherwise False
        """
        return self.is_in_ref(self.get_xi(x))

    def project_quantity(self, x: np.ndarray, q: np.ndarray):
        """
        Projects the quantity from the nodal quantities q to the position x

        Args:
            x: Point in space
            q: Nodal values of the quantities

        Returns:
            q_x: Quantity at x projected with the shape functions
        """

        return self.project_quantity_xi(self.get_xi(x), q)

    def project_quantity_xi(self, xi: np.ndarray, q: np.ndarray):
        """
        Projects the quantity from the nodal quantities q to the local position in reference coordinates xi

        Args:
            xi: Point in the element in reference coordinates
            q: Nodal values of the quantities

        Returns:
            q_x: Quantity at xi projected with the shape functions
        """
        shapefcns = self.shape_fcns(xi)

        if q.shape[0] != self.get_num_nodes():
            raise ValueError(
                "Expecting nodal values of the quantities, expected {0} got {1}".format(
                    self.get_num_nodes(), q.shape[0]
                )
            )

        return np.dot(shapefcns, q)

    def integrate(self, integrand, numgp) -> np.ndarray:
        """
        Integrates the integrand over the element

        Args:
            integrand: integrand as a function of x
            numgp: Number of integration points to be used
        """
        raise NotImplementedError(
            "Integration is currently not implemented for all elements"
        )

    @classmethod
    def num_nodes_by_shape(cls, shape: str):
        """
        Returns the number of nodes from the shapes

        Args:
            shape: shape as used in baci .dat file format

        Returngs:
            int: number of nodes
        """
        shape_dict = {
            "VERTEX1": 1,
            "TET4": 4,
            "TET10": 10,
            "PYRAMID5": 5,
            "HEX8": 8,
            "HEX20": 20,
            "HEX27": 27,
            "WEDGE6": 6,
            "QUAD4": 4,
            "TRI3": 3,
            "TRI6": 6,
            "LINE2": 2,
            "LINE3": 3,
            "QUAD8": 8,
            "QUAD9": 9,
        }

        if shape not in shape_dict:
            raise ValueError("Element of shape {0} is unknown".format(shape))

        return shape_dict[shape]

    @classmethod
    def shape_fcns(cls, xi):
        """
        Returns the value of the shape functions at the local coordinate xi
        """
        raise NotImplementedError("This element has not implemented shape functions")

    @classmethod
    def int_points(cls, num_points):
        """
        Returns the position of the integration points for a given number of integration points
        """
        raise NotImplementedError("This element has not implemented integration points")


class Element0D(Element):
    @classmethod
    def get_space_dim(cls):
        """
        Returns the number of space dimensions

        Returns:
            Number of space dimensions
        """
        return 0


class Element1D(Element):
    @classmethod
    def get_space_dim(cls):
        """
        Returns the number of space dimensions

        Returns:
            Number of space dimensions
        """
        return 1


class Element2D(Element):
    @classmethod
    def get_space_dim(cls):
        """
        Returns the number of space dimensions

        Returns:
            Number of space dimensions
        """
        return 2


class Element3D(Element):
    @classmethod
    def get_space_dim(cls):
        """
        Returns the number of space dimensions

        Returns:
            Number of space dimensions
        """
        return 3


class ElementTri(Element2D):
    @classmethod
    def int_points(cls, num_points: int) -> np.ndarray:
        """
        Returns the position of the integration points for a given number of integration points
        """
        if num_points == 1:
            return np.array([[1.0 / 3.0, 1.0 / 3.0]])
        if num_points == 3:
            return np.array(
                [[1.0 / 6.0, 1.0 / 6.0], [2.0 / 3.0, 1.0 / 6.0], [1.0 / 6.0, 2.0 / 3.0]]
            )
        raise RuntimeError(
            "The number of integration points provided ({0}) is not supported for TRI elements".format(
                num_points
            )
        )

    @classmethod
    def int_weights(cls, num_points: int) -> np.ndarray:
        if num_points == 1:
            return np.array([0.5])
        elif num_points == 3:
            return np.array([1.0 / 6.0, 1.0 / 6.0, 1.0 / 6.0])
        raise RuntimeError(
            "The number of integration points provided ({0}) is not supported for TRI elements".format(
                num_points
            )
        )


class ElementQuad(Element2D):
    @classmethod
    def int_weights(cls, num_points: int) -> np.ndarray:
        """
        Returns the integration weights for the integration points
        """
        if num_points == 1:
            return np.array([2])
        elif num_points == 4:
            return np.array([1.0, 1.0, 1.0, 1.0])

        raise RuntimeError(
            "The number of integration points provided ({0}) is not supported for TET elements".format(
                num_points
            )
        )

    @classmethod
    def int_points(cls, num_points: int) -> np.ndarray:
        """
        Returns the position of the integration points for a given number of integration points
        """
        if num_points == 1:
            return np.array([[0.0, 0.0]])
        if num_points == 4:
            return np.array(
                [
                    [-0.5773502691896, -0.5773502691896],
                    [0.5773502691896, -0.5773502691896],
                    [-0.5773502691896, 0.5773502691896],
                    [0.5773502691896, 0.5773502691896],
                ]
            )
        raise RuntimeError(
            "The number of integration points provided ({0}) is not supported for TET elements".format(
                num_points
            )
        )


class ElementTet(Element3D):
    @classmethod
    def is_in_ref(cls, xi, include_boundary=True):
        if include_boundary:

            def lop(x, y):
                return x >= y

        else:

            def lop(x, y):
                return x > y

        if not all([lop(xi[i], 0) for i in range(3)]):
            return False

        if not lop(1, xi[0]):
            return False

        if not lop(1 - xi[0], xi[1]):
            return False

        if not lop(1 - xi[0] - xi[1], xi[2]):
            return False

        return True

    @classmethod
    def int_weights(cls, num_points):
        """
        Returns the integration weights for the integration points
        """
        if num_points == 1:
            return np.array([1.0 / 6.0])
        elif num_points == 4:
            return np.array(
                [1.0 / 6.0 / 4.0, 1.0 / 6.0 / 4.0, 1.0 / 6.0 / 4.0, 1.0 / 6.0 / 4.0]
            )

    @classmethod
    def int_points(cls, num_points):
        """
        Returns the position of the integration points for a given number of integration points
        """
        if num_points == 1:
            return np.array([0.25, 0.25, 0.25])
        if num_points == 4:
            palpha = (5.0 + 3.0 * math.sqrt(5.0)) / 20.0
            pbeta = (5.0 - math.sqrt(5.0)) / 20.0

            return np.array(
                [
                    [pbeta, pbeta, pbeta],
                    [palpha, pbeta, pbeta],
                    [pbeta, palpha, pbeta],
                    [pbeta, pbeta, palpha],
                ]
            )
        raise RuntimeError(
            "The number of integration points provided ({0}) is not supported for TET elements".format(
                num_points
            )
        )


class ElementHex(Element3D):
    @classmethod
    def is_in_ref(cls, xi, include_boundary=True):
        if include_boundary:

            def lop(x, y):
                return x >= y

        else:

            def lop(x, y):
                return x > y

        if not all([lop(xi[i], -1.0) for i in range(3)]):
            return False

        if not all([lop(1.0, xi[i]) for i in range(3)]):
            return False

        return True

    @classmethod
    def int_points(cls, num_points):
        """
        Returns the position of the integration points for a given number of integration points
        """
        if num_points == 8:
            xi = 1.0 / math.sqrt(3.0)

            return (
                np.array(
                    [
                        [-1, -1, -1],
                        [1, -1, -1],
                        [1, 1, -1],
                        [-1, 1, -1],
                        [-1, -1, 1],
                        [1, -1, 1],
                        [1, 1, 1],
                        [-1, 1, 1],
                    ]
                )
                * xi
            )
        raise RuntimeError(
            "The number of integration points provided ({0}) is not supported for HEX elements".format(
                num_points
            )
        )

    @classmethod
    def int_weights(cls, num_points):
        """
        Returns the integration weights for the integration points
        """
        if num_points == 8:
            return np.ones(8)



import os
import unittest

import lnmmeshio

script_dir = os.path.dirname(os.path.realpath(__file__))


class TestMeshio2Discretization(unittest.TestCase):
    def setUp(self):
        pass

    def test_get_dsurf_elements(self):
        dis: lnmmeshio.Discretization = lnmmeshio.read(
            os.path.join(script_dir, "data", "dummy2.dat")
        )

        surface_elements = dis.get_dsurf_elements(0)

        self.assertEqual(len(surface_elements), 1)
        self.assertIsInstance(surface_elements[0], lnmmeshio.Tri3)
        self.assertListEqual(
            sorted([n.id for n in surface_elements[0].nodes]), [0, 1, 2]
        )

    def test_get_dline_elements(self):
        dis: lnmmeshio.Discretization = lnmmeshio.read(
            os.path.join(script_dir, "data", "dummy2.dat")
        )

        line_elements = dis.get_dline_elements(0)

        self.assertEqual(len(line_elements), 1)
        self.assertIsInstance(line_elements[0], lnmmeshio.Line2)
        self.assertListEqual(sorted([n.id for n in line_elements[0].nodes]), [0, 1])

    def test_get_nodes(self):
        dis: lnmmeshio.Discretization = lnmmeshio.read(
            os.path.join(script_dir, "data", "dummy2.dat")
        )
        dis.compute_ids(zero_based=False)

        # test scalar
        self.assertListEqual([n.id for n in dis.pointnodesets[0]], [1])
        self.assertListEqual(sorted([n.id for n in dis.linenodesets[0]]), [1, 2])
        self.assertListEqual(sorted([n.id for n in dis.surfacenodesets[0]]), [1, 2, 3])
        self.assertListEqual(
            sorted([n.id for n in dis.volumenodesets[0]]), [1, 2, 3, 4]
        )

        # test list
        self.assertListEqual([n.id for n in dis.pointnodesets[0]], [1])
        self.assertListEqual(sorted([n.id for n in dis.linenodesets[0]]), [1, 2])
        self.assertListEqual(sorted([n.id for n in dis.surfacenodesets[0]]), [1, 2, 3])
        self.assertListEqual(
            sorted([n.id for n in dis.volumenodesets[0]]), [1, 2, 3, 4]
        )



#
# Maintainer: Amadeus Gebauer
#

import os
import sys
from typing import IO, Dict, List, Optional, Union

import numpy as np
from meshio import Mesh
from meshio import read as _meshioread
from meshio import write as _meshiowrite

from . import element, ensightio, ioutils, mimics_stlio, node, nodeset
from .discretization import Discretization
from .element.element import (
    Element,
    Element1D,
    Element2D,
    Element3D,
    ElementHex,
    ElementQuad,
    ElementTet,
    ElementTri,
)
from .element.element_container import ElementContainer
from .element.hex8 import Hex8
from .element.hex20 import Hex20
from .element.hex27 import Hex27
from .element.line2 import Line2
from .element.line3 import Line3
from .element.pyramid5 import Pyramid5
from .element.quad4 import Quad4
from .element.quad8 import Quad8
from .element.quad9 import Quad9
from .element.tet4 import Tet4
from .element.tet10 import Tet10
from .element.tri3 import Tri3
from .element.tri6 import Tri6
from .element.vertex import Vertex
from .element.wedge6 import Wedge6
from .fiber import Fiber
from .node import Node
from .nodeset import LineNodeset, PointNodeset, SurfaceNodeset, VolumeNodeset

__TYPE_LEGACY_DAT = 1
__TYPE_FOUR_C_YAML = 2
__TYPE_CASE = 3
__TYPE_CASE_ASCII = 5
__TYPE_MIMICS_STL = 4
__TYPE_OTHER = 0


def _get_type(filename: str, file_format: Optional[str] = None) -> int:
    if not file_format:
        # deduct file format from extension
        _, extension = os.path.splitext(filename)

        if extension == ".dat":
            return __TYPE_LEGACY_DAT
        elif extension == ".dis":
            return __TYPE_LEGACY_DAT
        elif extension == ".yaml" or extension == ".yml":
            return __TYPE_FOUR_C_YAML
        elif extension == ".case":
            return __TYPE_CASE
        elif extension == ".mstl":
            return __TYPE_MIMICS_STL
    elif file_format == "dat":
        return __TYPE_LEGACY_DAT
    elif file_format == "dis":
        return __TYPE_LEGACY_DAT
    elif file_format == "yaml" or file_format == "yml":
        return __TYPE_FOUR_C_YAML
    elif file_format == "case" or file_format == "ensight":
        return __TYPE_CASE
    elif file_format == "case_ascii":
        return __TYPE_CASE_ASCII
    elif file_format == "mimicsstl":
        return __TYPE_MIMICS_STL

    return __TYPE_OTHER


def read(
    filename: str, file_format: Optional[str] = None, out: bool = True
) -> Discretization:
    """
    Reads an unstructured mesh with added data

    Args:
        filename: The file to read from
        file_format: The file format of the file

    Returns:
        Discretization: Returns the discretization in 4C format
    """
    assert isinstance(filename, str)

    ftype: int = _get_type(filename, file_format=file_format)

    if ftype == __TYPE_LEGACY_DAT:
        # this is a legacy 4C dat file format
        with open(filename, "r") as f:
            return read_legacy_four_c_dat(f, out=out)
    elif ftype == __TYPE_FOUR_C_YAML:
        # this is a 4C yaml file format
        with open(filename, "r") as f:
            return read_four_c_yaml(f, out=out)
    elif ftype == __TYPE_CASE:
        # this is ensight gold file format
        raise NotImplementedError("Case file reading is not implemented yet")
    elif ftype == __TYPE_MIMICS_STL:
        return from_mesh(mimics_stlio.read(filename))
    else:
        # this maybe is a file format supported by meshio
        return from_mesh(_meshioread(filename, file_format=file_format))


def read_legacy_four_c_dat(input_stream: IO, out: bool = True) -> Discretization:
    sections = ioutils.read_dat_sections(input_stream)

    return Discretization.read(sections, out=out)


def read_four_c_yaml(input_stream: IO, out: bool = True) -> Discretization:
    import yaml

    return Discretization.read(yaml.safe_load(input_stream), out=out)


def write(
    filename: str,
    dis: Discretization,
    file_format=None,
    override=True,
    out=True,
) -> None:
    """
    Writes an dat file with head and discretization

    Args:
        filename: The file to read from
        file_format: The file format of the file
        override: Flag, whether existing files should be overriden (dangerous)
    """
    assert isinstance(filename, str)

    if not override and os.path.exists(filename):
        raise FileExistsError("The file already exists")

    ftype: int = _get_type(filename, file_format=file_format)

    if ftype == __TYPE_LEGACY_DAT:
        # this is a legacy 4C dat discretization file format
        with open(filename, "w") as f:
            dis.write_legacy_dat(f, out=out)
    elif ftype == __TYPE_FOUR_C_YAML:
        # this is a BACI discretization file format
        with open(filename, "w") as f:
            dis.write_yaml(f, out=out)
    elif ftype == __TYPE_CASE:
        # this is ensight gold file format
        ensightio.write_case(filename, dis, out=out)
    elif ftype == __TYPE_CASE_ASCII:
        # this is ensight gold file format
        ensightio.write_case(filename, dis, out=out, binary=False)
    elif ftype == __TYPE_MIMICS_STL:
        raise NotImplementedError("Writing in Mimics stl is currently not supported")
    else:
        write_mesh(filename, to_mesh(dis), file_format=file_format)


def read_mesh(filename: str, file_format: Optional[str] = None) -> Mesh:
    """
    Reads an unstructured mesh with added data into the meshio raw format

    Args:
        filename: The file to read from
        file_format: The file format of the file

    Returns:
        Mesh: Returns the discretization in 4C format
    """
    assert isinstance(filename, str)

    ftype: int = _get_type(filename, file_format=file_format)

    if ftype == __TYPE_LEGACY_DAT:
        # this is a 4C file format
        return to_mesh(read(filename, file_format=file_format))
    elif ftype == __TYPE_FOUR_C_YAML:
        # this is a 4C file format
        return to_mesh(read(filename, file_format=file_format))
    elif ftype == __TYPE_CASE:
        # this is ensight gold file format
        raise NotImplementedError("Case file reading into mesh is not implemented yet")
    elif ftype == __TYPE_MIMICS_STL:
        return mimics_stlio.read(filename)
    else:
        # this maybe is a file format supported by meshio
        return _meshioread(filename, file_format=file_format)


def write_mesh(
    filename: str,
    mesh: Mesh,
    file_format: Optional[str] = None,
    override: bool = True,
    **kwargs,
) -> None:
    """
    Writes an unstructured mesh with added data from meshio raw data

    Args:
        filename: The file to read from
        mesh: meshio raw mesh data
        file_format: The file format of the file
        override: Flag, whether existing files should be overriden (dangerous)
    """
    assert isinstance(filename, str)

    ftype: int = _get_type(filename, file_format=file_format)

    if not override and os.path.isfile(filename):
        raise FileExistsError("The file already exists")

    if ftype == __TYPE_LEGACY_DAT or ftype == __TYPE_FOUR_C_YAML:
        dis = from_mesh(mesh)
        write(
            filename,
            dis,
            file_format=file_format,
        )
    elif ftype == __TYPE_CASE:
        # this is ensight gold file format
        raise NotImplementedError("Case file writing into mesh is not implemented yet")
    elif ftype == __TYPE_MIMICS_STL:
        mimics_stlio.write(filename, mesh)
    else:
        _meshiowrite(filename, mesh, file_format, **kwargs)


def read_sections(filename: str) -> Dict[str, List[str]]:
    """
    Reads the sections of a dat file and returns a dictionary with the section title as key and the
    lines as an array

    Args:
        filename: Path to the datfile

    Returns:
        dict: keys are the section names and value is a list of the lines
    """
    with open(filename, "r") as f:
        sections = ioutils.read_dat_sections(f)

    return sections


def from_mesh(mesh: Mesh) -> Discretization:
    """
    Converts a meshio.Mesh to a discretization
    """
    from . import meshio_to_discretization

    return meshio_to_discretization.mesh2Discretization(mesh)


def to_mesh(mesh: Discretization) -> Mesh:
    """
    Converts a discretization to a meshio.Mesh
    """
    from . import meshio_to_discretization

    return meshio_to_discretization.discretization2mesh(mesh)


def to_pyvista(dis: Discretization):
    """
    Converts a discretization to a pyvista unstructured grid
    """
    import pyvista

    return pyvista.from_meshio(to_mesh(dis))



import os
import unittest

from lnmmeshio import ElementContainer

script_dir = os.path.dirname(os.path.realpath(__file__))


class TestEleContainer(unittest.TestCase):
    def setUp(self):
        pass

    def test_elecontainer(self):
        c = ElementContainer()

        self.assertFalse(ElementContainer.TypeStructure in c)
        self.assertFalse(ElementContainer.TypeFluid in c)
        self.assertFalse(ElementContainer.TypeALE in c)
        self.assertFalse(ElementContainer.TypeTransport in c)
        self.assertFalse(ElementContainer.TypeThermo in c)
        self.assertFalse(ElementContainer.TypeArtery in c)

        self.assertRaises(KeyError, lambda: c[ElementContainer.TypeStructure])
        self.assertRaises(KeyError, lambda: c[ElementContainer.TypeFluid])
        self.assertRaises(KeyError, lambda: c[ElementContainer.TypeALE])
        self.assertRaises(KeyError, lambda: c[ElementContainer.TypeTransport])
        self.assertRaises(KeyError, lambda: c[ElementContainer.TypeThermo])
        self.assertRaises(KeyError, lambda: c[ElementContainer.TypeArtery])
        self.assertRaises(KeyError, lambda: c["doesnotexist"])
        with self.assertRaises(KeyError) as _:
            c["doesnotexist"] = 1

        c[ElementContainer.TypeStructure] = [1]
        self.assertListEqual(c[ElementContainer.TypeStructure], [1])
        self.assertListEqual(c.structure, [1])

        c[ElementContainer.TypeFluid] = [2]
        self.assertListEqual(c[ElementContainer.TypeFluid], [2])
        self.assertListEqual(c.fluid, [2])

        c[ElementContainer.TypeALE] = [3]
        self.assertListEqual(c[ElementContainer.TypeALE], [3])
        self.assertListEqual(c.ale, [3])

        c[ElementContainer.TypeTransport] = [4]
        self.assertListEqual(c[ElementContainer.TypeTransport], [4])
        self.assertListEqual(c.transport, [4])

        c[ElementContainer.TypeThermo] = [5]
        self.assertListEqual(c[ElementContainer.TypeThermo], [5])
        self.assertListEqual(c.thermo, [5])

        c[ElementContainer.TypeArtery] = [5]
        self.assertListEqual(c[ElementContainer.TypeArtery], [5])
        self.assertListEqual(c.artery, [5])

        self.assertTrue(ElementContainer.TypeStructure in c)
        self.assertTrue(ElementContainer.TypeFluid in c)
        self.assertTrue(ElementContainer.TypeALE in c)
        self.assertTrue(ElementContainer.TypeTransport in c)
        self.assertTrue(ElementContainer.TypeThermo in c)
        self.assertTrue(ElementContainer.TypeArtery in c)



from typing import Dict, List, Optional

from tqdm import tqdm

from ..ioutils import write_title
from ..node import Node
from .element import Element
from .parse_element import parse as parse_ele


class ElementContainer:
    """
    Class holding all elements in different categories. Current implemented categories are

    ElementContainer.structure
    ElementContainer.fluid
    ElementContainer.ale
    ElementContainer.transport
    ElementContainer.thermo
    ElementContainer.artery
    """

    TypeStructure: str = "structure"
    TypeFluid: str = "fluid"
    TypeALE: str = "ale"
    TypeTransport: str = "transport"
    TypeThermo: str = "thermo"
    TypeArtery: str = "artery"

    def __init__(self):
        """
        Initialize element contained
        """
        self.structure: Optional[List[Element]] = None
        self.fluid: Optional[List[Element]] = None
        self.ale: Optional[List[Element]] = None
        self.transport: Optional[List[Element]] = None
        self.thermo: Optional[List[Element]] = None
        self.artery: Optional[List[Element]] = None

    def get_num_structure(self) -> int:
        """
        Returns the number of structural elements in the discretization

        Returns:
            int: Number of structural parameters in the discretization
        """
        if self.structure is None:
            return 0

        return len(self.structure)

    def get_num_fluid(self) -> int:
        """
        Returns the number of fluid elements in the discretization

        Returns:
            int: Number of fluid parameters in the discretization
        """
        if self.fluid is None:
            return 0

        return len(self.fluid)

    def get_num_ale(self) -> int:
        """
        Returns the number of ale elements in the discretization

        Returns:
            int: Number of ale parameters in the discretization
        """
        if self.ale is None:
            return 0

        return len(self.ale)

    def get_num_transport(self) -> int:
        """
        Returns the number of transport elements in the discretization

        Returns:
            int: Number of transport parameters in the discretization
        """
        if self.transport is None:
            return 0

        return len(self.transport)

    def get_num_thermo(self) -> int:
        """
        Returns the number of thermo elements in the discretization

        Returns:
            int: Number of thermo parameters in the discretization
        """
        if self.thermo is None:
            return 0

        return len(self.thermo)

    def get_num_artery(self) -> int:
        """
        Returns the number of artery elements in the discretization

        Returns:
            int: Number of artery parameters in the discretization
        """
        if self.artery is None:
            return 0

        return len(self.artery)

    def write(self, dest):
        """
        Write the corresponding sections into the stream like variable dest

        Args:
            dest: Stream like variable, where to write the corresponding sections
        """
        if self.structure is not None:
            ElementContainer.__write_section(dest, "STRUCTURE ELEMENTS", self.structure)

        if self.fluid is not None:
            ElementContainer.__write_section(dest, "FLUID ELEMENTS", self.fluid)

        if self.ale is not None:
            ElementContainer.__write_section(dest, "ALE ELEMENTS", self.ale)

        if self.transport is not None:
            ElementContainer.__write_section(dest, "TRANSPORT ELEMENTS", self.transport)

        if self.thermo is not None:
            ElementContainer.__write_section(dest, "THERMO ELEMENTS", self.thermo)

        if self.artery is not None:
            ElementContainer.__write_section(dest, "ARTERY ELEMENTS", self.artery)

    def get_sections(self, out=True):
        """
        Returns a dict of sections with the corresponding lines
        """
        d = {}
        for key, elements in self.items():
            d[ElementContainer.get_section_name(key)] = (
                ElementContainer.__get_section_lines(elements, out=out)
            )

        return d

    def values(self):
        """
        Returns a list of List[Elements] for the different element types

        Returns:
            List[List[Element]]
        """
        ele_list = []

        if self.structure is not None:
            ele_list.append(self.structure)
        if self.fluid is not None:
            ele_list.append(self.fluid)
        if self.ale is not None:
            ele_list.append(self.ale)
        if self.transport is not None:
            ele_list.append(self.transport)
        if self.thermo is not None:
            ele_list.append(self.thermo)
        if self.artery is not None:
            ele_list.append(self.artery)

        return ele_list

    def items(self):
        """
        Returns a list of tuples with key and values for each element type

        Returns:
            List[Tuple[str, List[Element]]]
        """
        return zip(self.keys(), self.values())

    def keys(self):
        """
        Returns a list of keys of the elemet types

        Returns:
            List[str]
        """
        ele_keys = []

        if self.structure is not None:
            ele_keys.append(self.TypeStructure)
        if self.fluid is not None:
            ele_keys.append(self.TypeFluid)
        if self.ale is not None:
            ele_keys.append(self.TypeALE)
        if self.transport is not None:
            ele_keys.append(self.TypeTransport)
        if self.thermo is not None:
            ele_keys.append(self.TypeThermo)
        if self.artery is not None:
            ele_keys.append(self.TypeArtery)

        return ele_keys

    def __getitem__(self, key):
        """
        Returns the list of elements belonging to a type of discretization

        Args:
            key: str Type of field

        Returns:
            List[Element]
        """
        if key == self.TypeStructure and self.structure is not None:
            return self.structure
        elif key == self.TypeFluid and self.fluid is not None:
            return self.fluid
        elif key == self.TypeALE and self.ale is not None:
            return self.ale
        elif key == self.TypeTransport and self.transport is not None:
            return self.transport
        elif key == self.TypeThermo and self.thermo is not None:
            return self.thermo
        elif key == self.TypeArtery and self.artery is not None:
            return self.artery

        raise KeyError("Key not found: {0}".format(key))

    def __contains__(self, key):
        """
        Returns the a boolean, whether the field type is available

        Args:
            key: str Type of field

        Returns:
            bool
        """
        if key == self.TypeStructure and self.structure is not None:
            return True
        elif key == self.TypeFluid and self.fluid is not None:
            return True
        elif key == self.TypeALE and self.ale is not None:
            return True
        elif key == self.TypeTransport and self.transport is not None:
            return True
        elif key == self.TypeThermo and self.thermo is not None:
            return True
        elif key == self.TypeArtery and self.artery is not None:
            return True

        return False

    def __setitem__(self, key, value):
        """
        Returns the list of elements belonging to a type of discretization

        Args:
            key: str Type of field
            value: List[Element] List of elements
        """
        if key == self.TypeStructure:
            self.structure = value
        elif key == self.TypeFluid:
            self.fluid = value
        elif key == self.TypeALE:
            self.ale = value
        elif key == self.TypeTransport:
            self.transport = value
        elif key == self.TypeThermo:
            self.thermo = value
        elif key == self.TypeArtery:
            self.artery = value
        else:
            raise KeyError("Key not found: {0}".format(key))

    @staticmethod
    def __write_section(dest, section_name: str, elements: List[Element]):
        """
        Writes the element section with section title and elements. If elements is None,
        the section is not written at all

        Args:
            dest: stream like variable to write
            section_name: string of the section heart
            elements: List of elements
        """
        if elements is not None:
            write_title(dest, section_name)
            for line in ElementContainer.__get_section_lines(elements):
                dest.write("{0}\n".format(line))

    @staticmethod
    def __get_section_lines(elements: List[Element], out=True):
        """
        Writes the elements into an list of lines

        Args:
            elements: List of elements
        """
        lines = []
        if elements is not None:
            for ele in tqdm(elements, disable=not out, desc="Write Element"):
                lines.append(ele.get_line())

        return lines

    @staticmethod
    def read_element_sections(
        sections: Dict[str, List[str]], nodes: List[Node], out=False
    ):
        elec = ElementContainer()
        # read elements
        if "STRUCTURE ELEMENTS" in sections:
            elec.structure = ElementContainer.__read_elements(
                nodes,
                sections["STRUCTURE ELEMENTS"],
                out=out,
                fieldtype="Structural elements",
            )

        if "FLUID ELEMENTS" in sections:
            elec.fluid = ElementContainer.__read_elements(
                nodes, sections["FLUID ELEMENTS"], out=out, fieldtype="Fluid elements"
            )

        if "ALE ELEMENTS" in sections:
            elec.ale = ElementContainer.__read_elements(
                nodes, sections["ALE ELEMENTS"], out=out, fieldtype="ALE elements"
            )

        if "TRANSPORT ELEMENTS" in sections:
            elec.transport = ElementContainer.__read_elements(
                nodes,
                sections["TRANSPORT ELEMENTS"],
                out=out,
                fieldtype="Transport elements",
            )

        if "THERMO ELEMENTS" in sections:
            elec.thermo = ElementContainer.__read_elements(
                nodes, sections["THERMO ELEMENTS"], out=out, fieldtype="Thermo elements"
            )

        if "ARTERY ELEMENTS" in sections:
            elec.thermo = ElementContainer.__read_elements(
                nodes, sections["ARTERY ELEMENTS"], out=out, fieldtype="Artery elements"
            )

        return elec

    @staticmethod
    def __read_elements(nodes: List[Node], lines: List[str], out=False, fieldtype=None):
        """
        Static method that reads the list of elements

        Args:
            nodes: List of nodes (order is important: first node in list must be the one with id 1)
            lines: List of string that represent the lines of the corresponding element section

        Returns:
            List of elements
        """
        eles: List[Element] = []

        if fieldtype is None:
            fieldtype = "Elements"

        for line in tqdm(lines, disable=not out, desc=fieldtype):
            ele = parse_ele(line, nodes)

            if ele is None:
                continue

            eles.append(ele)

            if ele.id is None:
                raise RuntimeError("Id of the element is None")

            # safety check for integrity of the dat file:
            # ids must increase continuously by +1 from first id
            first_id: int = 0
            id_offset: int = 0
            if len(eles) == 1:
                first_id = eles[0].id if eles[0].id is not None else 1
                id_offset = first_id - 1
            if int(ele.id) - id_offset != len(eles):
                raise RuntimeError(
                    "Element ids in dat file have a gap at {0}!={1}!".format(
                        ele.id, len(eles)
                    )
                )

        return eles

    @staticmethod
    def get_section_name(fieldtype):
        if fieldtype == ElementContainer.TypeStructure:
            return "STRUCTURE ELEMENTS"
        if fieldtype == ElementContainer.TypeFluid:
            return "FLUID ELEMENTS"
        if fieldtype == ElementContainer.TypeALE:
            return "ALE ELEMENTS"
        if fieldtype == ElementContainer.TypeTransport:
            return "TRANSPORT ELEMENTS"
        if fieldtype == ElementContainer.TypeThermo:
            return "THERMO ELEMENTS"
        if fieldtype == ElementContainer.TypeArtery:
            return "ARTERY ELEMENTS"



import filecmp
import os
import unittest

import lnmmeshio
import numpy as np

script_dir = os.path.dirname(os.path.realpath(__file__))


class TestEnsight(unittest.TestCase):
    def setUp(self):
        if not os.path.isdir(os.path.join(script_dir, "tmp")):
            os.makedirs(os.path.join(script_dir, "tmp"))

    def test_write_ensight_ascii(self):
        dis: lnmmeshio.Discretization = lnmmeshio.read(
            os.path.join(script_dir, "data", "dummy.dat")
        )
        dis.compute_ids(zero_based=False)

        for elelist in dis.elements.values():
            for ele in elelist:
                ele.data["material"] = np.array(float(" ".join(ele.options["MAT"])))

        for n in dis.nodes:
            n.data["id"] = np.array([n.id])

        # write ensight
        lnmmeshio.ensightio.write_case(
            os.path.join(script_dir, "tmp", "ensight_ascii.case"),
            dis,
            binary=False,
            override=True,
        )

        # how to check?

    def test_write_ensight_binary(self):
        dis: lnmmeshio.Discretization = lnmmeshio.read(
            os.path.join(script_dir, "data", "dummy.dat")
        )
        dis.compute_ids(zero_based=False)

        for elelist in dis.elements.values():
            for ele in elelist:
                ele.data["material"] = np.array(float(" ".join(ele.options["MAT"])))

        for n in dis.nodes:
            n.data["id"] = np.array([n.id])

        # write ensight
        lnmmeshio.ensightio.write_case(
            os.path.join(script_dir, "tmp", "ensight_binary.case"),
            dis,
            binary=True,
            override=True,
        )

        # how to check?

    def test_write_float(self):
        ref_file = os.path.join(script_dir, "tmp", "ref.tmp")
        res_file = os.path.join(script_dir, "tmp", "res.tmp")

        arr = np.random.rand(10, 20)

        # write ref binary file
        with open(ref_file, "wb") as f:
            np.ravel(arr, "F").astype("<f").tofile(f)

        # test binary
        with open(res_file, "wb") as f:
            lnmmeshio.ioutils.ens_write_floats(f, arr, binary=True)

        self.assertTrue(filecmp.cmp(res_file, ref_file))

        # write ref ascii file
        with open(ref_file, "w") as f:
            for v in np.ravel(arr, "F").astype("<f"):
                if v >= 0:
                    f.write(" ")

                f.write("{0:10.5e}\n".format(v))

        # test binary
        with open(res_file, "w") as f:
            lnmmeshio.ioutils.ens_write_floats(f, arr, binary=False)

        self.assertTrue(filecmp.cmp(res_file, ref_file))

    def test_write_ints(self):
        ref_file = os.path.join(script_dir, "tmp", "ref.tmp")
        res_file = os.path.join(script_dir, "tmp", "res.tmp")

        arr = np.arange(20).reshape((2, -1))

        # write ref binary file
        with open(ref_file, "wb") as f:
            np.ravel(arr).astype("<i").tofile(f)

        # test binary
        with open(res_file, "wb") as f:
            lnmmeshio.ioutils.ens_write_ints(f, arr, binary=True)

        self.assertTrue(filecmp.cmp(res_file, ref_file))

    def tearDown(self):
        pass
        # delete tmp folder
        # shutil.rmtree(os.path.join(script_dir, 'tmp'))



"""
This file is mainly based on the stl-reader by meshio (https://github.com/nschloe/meshio).

Copyright (c) 2015-2021 Nico Schlömer et al.

I/O for the STL format, cf.
<https://en.wikipedia.org/wiki/STL_(file_format)>.
"""

import io
import logging
import struct

import numpy
from meshio import Mesh


def read(filename):
    """Reads a Gmsh msh file."""
    with open(filename, "rb") as f:
        out = read_buffer(f)
    return out


def read_buffer(f):
    data = numpy.frombuffer(f.read(5), dtype=numpy.uint8)
    if "".join([chr(item) for item in data]) == "solid":
        raise RuntimeError(
            "This is the Mimics stl format, which should be binary, not ascii"
        )

    # binary: read and discard 75 more bytes
    f.read(75)
    return _read_binary(f)


def data_from_facets(facets):
    # Now, all facets contain the point coordinate. Try to identify individual
    # points and build the data arrays.
    pts = numpy.concatenate(facets)

    # TODO equip `unique()` with a tolerance
    # Use return_index so we can use sort on `idx` such that the order is
    # preserved; see <https://stackoverflow.com/a/15637512/353337>.
    _, idx, inv = numpy.unique(pts, axis=0, return_index=True, return_inverse=True)
    k = numpy.argsort(idx)
    points = pts[idx[k]]
    inv_k = numpy.argsort(k)
    cells = [("triangle", inv_k[inv].reshape(-1, 3))]
    return points, cells


def _read_binary(f):
    # read the first uint32 byte to get the number of triangles
    data = numpy.frombuffer(f.read(4), dtype=numpy.uint32)
    num_triangles = data[0]

    facets = []
    facet_data = []
    for _ in range(num_triangles):
        # discard the normal
        f.read(12)
        facets.append(numpy.frombuffer(f.read(36), dtype=numpy.float32).reshape(-1, 3))
        # discard the attribute byte count
        facet_data.append(int(struct.unpack("<h", f.read(2))[0]))

    points, cells = data_from_facets(numpy.array(facets))

    cell_data = {"medit:ref": [numpy.array(facet_data)]}
    return Mesh(points, cells, cell_data=cell_data)


def write(filename, mesh):
    assert (
        len(mesh.cells) == 1 and mesh.cells[0].type == "triangle"
    ), "STL can only write triangle cells."

    if mesh.points.shape[1] == 2:
        logging.warning(
            "STL requires 3D points, but 2D points given. "
            "Appending 0 third component."
        )
        mesh.points = numpy.column_stack(
            [mesh.points[:, 0], mesh.points[:, 1], numpy.zeros(mesh.points.shape[0])]
        )

    _write_binary(filename, mesh.points, mesh.cells, cell_data=mesh.cell_data)

    return


def _compute_normals(pts):
    normals = numpy.cross(pts[:, 1] - pts[:, 0], pts[:, 2] - pts[:, 0])
    nrm = numpy.sqrt(numpy.einsum("ij,ij->i", normals, normals))
    normals = (normals.T / nrm).T
    return normals


def _write_binary(filename, points, cells, cell_data=None):
    pts = None
    attrs = None
    for i, t in enumerate(cells):
        if t.type == "triangle":
            pts = points[t.data]

            if cell_data is not None and "medit:ref" in cell_data:
                attrs = cell_data["medit:ref"][i].astype(numpy.uint16)
            break

    if pts is None:
        raise RuntimeError("Could not find triangled")

    if attrs is None:
        attrs = numpy.zeros((len(pts)), dtype=numpy.uint16)

    normals = _compute_normals(pts)

    with open(filename, "wb") as fh:
        # 80 character header data containing default color for 3-matic
        byte_io = io.BytesIO()
        byte_io.write(
            b"COLOR=\xc0\xc0\xc0\x00,MATERIAL=\x80\x80\x80\xff\x80\x80\x80\xff\x00\x00\x00\xff\x00\x00\xa0A;SOLID\x00"
        )
        byte_io.write(b" " * (80 - byte_io.getbuffer().nbytes))

        fh.write(byte_io.getbuffer().tobytes())
        fh.write(numpy.uint32(len(cells[0].data)))
        for pt, normal, attr in zip(pts, normals, attrs):
            fh.write(normal.astype(numpy.float32))
            fh.write(pt.astype(numpy.float32))
            fh.write(attr)

    return



import os
import shutil
import tempfile
import unittest

import lnmmeshio
import numpy as np

script_dir = os.path.dirname(os.path.realpath(__file__))


class TestWriteData(unittest.TestCase):
    def test_write_element_data(self):
        d = lnmmeshio.Discretization()
        d.nodes = [
            lnmmeshio.Node(np.array([0.0, 0.0, 0.0])),
            lnmmeshio.Node(np.array([1.0, 0.0, 0.0])),
            lnmmeshio.Node(np.array([2.0, 0.0, 0.0])),
        ]
        d.elements.structure = [
            lnmmeshio.Line2("LINE", [d.nodes[0], d.nodes[1]]),
            lnmmeshio.Line2("LINE", [d.nodes[1], d.nodes[2]]),
        ]

        d.elements.structure[1].data["test"] = 1

        with tempfile.TemporaryDirectory() as tmp:
            filename = os.path.join(tmp, "test.vtu")
            lnmmeshio.write(filename, d)

            dis_read = lnmmeshio.read(filename)

            self.assertIsNotNone(dis_read.elements.structure)
            self.assertEqual(len(dis_read.elements.structure), 2)
            self.assertIn("test", dis_read.elements.structure[0].data)
            self.assertIn("test", dis_read.elements.structure[1].data)
            self.assertAlmostEqual(dis_read.elements.structure[0].data["test"], 0)
            self.assertAlmostEqual(dis_read.elements.structure[1].data["test"], 1)

    def test_write_node_data(self):
        d = lnmmeshio.Discretization()
        d.nodes = [
            lnmmeshio.Node(np.array([0.0, 0.0, 0.0])),
            lnmmeshio.Node(np.array([1.0, 0.0, 0.0])),
            lnmmeshio.Node(np.array([2.0, 0.0, 0.0])),
        ]
        d.elements.structure = [
            lnmmeshio.Line2("LINE", [d.nodes[0], d.nodes[1]]),
            lnmmeshio.Line2("LINE", [d.nodes[1], d.nodes[2]]),
        ]

        d.nodes[1].data["test"] = 1

        with tempfile.TemporaryDirectory() as tmp:
            filename = os.path.join(tmp, "test.vtu")
            lnmmeshio.write(filename, d)

            dis_read = lnmmeshio.read(filename)

            self.assertEqual(len(dis_read.nodes), 3)

            for i in range(3):
                self.assertIn("test", dis_read.nodes[i].data)
                self.assertAlmostEqual(dis_read.nodes[i].data["test"], int(i == 1))



from typing import List

import numpy as np

from ..node import Node
from .element import ElementTet
from .line2 import Line2
from .tri3 import Tri3


class Tet4(ElementTet):
    """
    Implementation of a tet4 element
    """

    ShapeName: str = "TET4"
    ShapeFunctionsN: np.ndarray = np.array(
        [[-1.0, -1.0, -1.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
    )
    ShapeFunctionsB: np.ndarray = np.array([1.0, 0.0, 0.0, 0.0])
    NodalReferenceCoordinates: np.ndarray = np.array(
        [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
    )

    def __init__(self, el_type: str, nodes: List[Node]):
        """
        Base constructor of a tet4 element
        """
        super(Tet4, self).__init__(el_type, Tet4.ShapeName, nodes)

        if len(nodes) != self.get_num_nodes():
            raise RuntimeError(
                "You tried to created a TET4 element with {0} nodes".format(len(nodes))
            )

    @classmethod
    def get_num_nodes(cls) -> int:
        """
        Get number of nodes of a TET4 element

        Returns:
            Number of nodes of a tet4 element = 4
        """
        return 4

    def get_faces(self) -> List[Tri3]:
        """
        Returns a list of faces of the tet4 element

        Returns:
            List of faces
        """
        face_node_ids = [[0, 1, 3], [1, 2, 3], [2, 0, 3], [0, 2, 1]]

        return [Tri3(None, [self.nodes[i] for i in nodes]) for nodes in face_node_ids]

    def get_edges(self) -> List[Line2]:
        """
        Returns the list of all edges

        Returns:
            List of edges
        """
        edge_node_ids = [[0, 1], [1, 2], [2, 0], [0, 3], [1, 3], [2, 3]]
        return [Line2(None, [self.nodes[i] for i in nodes]) for nodes in edge_node_ids]

    def get_xi(self, x):
        coords = np.transpose(np.array([n.coords for n in self.nodes]))
        N, b = Tet4.shape_fcns_mv()

        return np.linalg.solve(np.dot(coords, N), x - np.dot(coords, b))

    @staticmethod
    def shape_fcns(xi):
        """
        Returns the value of the shape functions at the local coordinate xi
        """
        N, b = Tet4.shape_fcns_mv()
        if len(xi.shape) == 2:
            b = b.reshape((4, 1))
        return np.matmul(N, xi) + b

    @staticmethod
    def shape_fcns_mv():
        return (
            Tet4.ShapeFunctionsN,
            Tet4.ShapeFunctionsB,
        )

    @staticmethod
    def shape_fcns_derivs(xi):
        """
        Returns the value of the derivatives of the shape functions with respect to the local coordinates at the local coordinate xi

        +-                                  -+
        |  dN_1 / dxi_1   dN_2 / dxi_1, ...  |
        |  dN_1 / dxi_2   dN_2 / dxi_2, ...  |
        |  dN_1 / dxi_3   dN_2 / dxi_3, ...  |
        +-                                  -+
        """
        return np.array(
            [
                [
                    -1,
                    1,
                    0,
                    0,
                ],
                [-1, 0, 1, 0],
                [
                    -1,
                    0,
                    0,
                    1,
                ],
            ]
        )

    @staticmethod
    def nodal_reference_coordinates():
        return Tet4.NodalReferenceCoordinates



from typing import List, Optional

import numpy as np

from ..node import Node
from .element import Element1D


class Line3(Element1D):
    """
    Implementation of a line3 element
    """

    ShapeName: str = "LINE3"

    def __init__(self, el_type: Optional[str], nodes: List[Node]):
        """
        Base constructor of a line3 element
        """
        super(Line3, self).__init__(el_type, Line3.ShapeName, nodes)

        if len(nodes) != self.get_num_nodes():
            raise RuntimeError(
                "You tried to created a LINE3 element with {0} nodes".format(len(nodes))
            )

    @classmethod
    def get_num_nodes(cls) -> int:
        """
        Get number of nodes of a LINE3 element

        Returns:
            Number of nodes of a line3 element = 3
        """
        return 3

    def get_faces(self) -> List:
        """
        A LINE3 element does not contain any faces

        Returns:
            List of faces
        """
        return []

    def get_edges(self) -> List["Line3"]:
        """
        Returns the list of all edges

        Returns:
            List of edges
        """
        return [self]

    @staticmethod
    def shape_fcns(xi):
        """
        Returns the value of the shape functions at the local coordinate xi
        """
        return np.array([xi * (xi - 1) / 2.0, xi * (xi + 1) / 2.0, (1 + xi) * (1 - xi)])



from typing import List

import numpy as np
from lnmmeshio.element.quad4 import Quad4
from lnmmeshio.element.tri3 import Tri3

from ..node import Node
from .element import Element2D, Element3D


class Pyramid5(Element3D):
    """
    Implementation of a PYRAMID5 element
    """

    ShapeName: str = "PYRAMID5"

    def __init__(self, el_type: str, nodes: List[Node]):
        """
        Base constructor of a Pyramid5 element
        """
        super(Pyramid5, self).__init__(el_type, Pyramid5.ShapeName, nodes)

        if len(nodes) != self.get_num_nodes():
            raise RuntimeError(
                "You tried to created a PYRAMID5 element with {0} nodes".format(
                    len(nodes)
                )
            )

    @classmethod
    def get_num_nodes(cls) -> int:
        """
        Get number of nodes of a PYRAMID5 element

        Returns:
            Number of nodes of a PYRAMID5 element = 5
        """
        return 5

    def get_faces(self) -> List[Element2D]:
        """
        Returns a list of faces of the Wedge6 element

        Returns:
            List of faces
        """
        face_node_ids = [[0, 1, 2, 3], [0, 1, 4], [1, 2, 4], [2, 3, 4], [3, 0, 4]]

        def gen_face(nodes):
            if len(nodes) == 3:
                return Tri3(None, nodes)
            else:
                return Quad4(None, nodes)

        return [gen_face([self.nodes[i] for i in nodes]) for nodes in face_node_ids]



from typing import List

import numpy as np

from ..node import Node
from .element import Element1D


class Line2(Element1D):
    """
    Implementation of a line2 element
    """

    ShapeName: str = "LINE2"

    def __init__(self, el_type: str, nodes: List[Node]):
        """
        Base constructor of a line2 element
        """
        super(Line2, self).__init__(el_type, Line2.ShapeName, nodes)

        if len(nodes) != self.get_num_nodes():
            raise RuntimeError(
                "You tried to created a LINE2 element with {0} nodes".format(len(nodes))
            )

    @classmethod
    def get_num_nodes(cls) -> int:
        """
        Get number of nodes of a LINE2 element

        Returns:
            Number of nodes of a line2 element = 2
        """
        return 2

    def get_faces(self) -> List["Line2"]:
        """
        A LINE2 element does not contain any faces

        Returns:
            List of faces
        """
        return []

    def get_edges(self) -> List["Line2"]:
        """
        Returns the list of all edges

        Returns:
            List of edges
        """
        return [self]

    @staticmethod
    def shape_fcns(xi):
        """
        Returns the value of the shape functions at the local coordinate xi
        """
        return np.array([(1 - xi) / 2.0, (1 + xi) / 2.0])

    @staticmethod
    def shape_fcns_derivs(xi):
        """
        Returns the value of the derivatives of the shape functions with respect to the local coordinates at the local coordinate xi

        +-                             -+
        |  dN_1 / dxi_1   dN_2 / dxi_1  |
        +-                             -+
        """
        return np.array([-0.5, 0.5])



import filecmp
import os
import shutil
import unittest

import lnmmeshio
import numpy as np

script_dir = os.path.dirname(os.path.realpath(__file__))


class TestMimicsStlIO(unittest.TestCase):
    def setUp(self):
        if not os.path.isdir(os.path.join(script_dir, "tmp")):
            os.makedirs(os.path.join(script_dir, "tmp"))

    def test_read_mimics_stl(self):
        mesh = lnmmeshio.read_mesh(
            os.path.join(script_dir, "data", "dummy.stl"), file_format="mimicsstl"
        )

        # check, whether everything is ok
        self.assertEqual(len(mesh.cells), 1)
        self.assertEqual(mesh.cells[0].type, "triangle")
        self.assertEqual(mesh.cells[0].data.shape, (18562, 3))
        self.assertEqual(mesh.points.shape, (9283, 3))

        # check, whether material is recognized
        self.assertEqual(np.sum(mesh.cell_data["medit:ref"][0] == 1), 10850)
        self.assertEqual(np.sum(mesh.cell_data["medit:ref"][0] == 2), 2006)
        self.assertEqual(np.sum(mesh.cell_data["medit:ref"][0] == 3), 5706)

    def test_write_mimics_stl(self):
        points = np.array(
            [
                [0, 0, 0],
                [10, 0, 0],
                [10, 10, 0],
                [0, 10, 0],
                [0, 0, 10],
                [10, 0, 10],
                [10, 10, 10],
                [0, 10, 10],
            ],
            dtype=float,
        )

        cells = []
        cells.append(
            (
                "triangle",
                np.array(
                    [
                        [0, 1, 4],
                        [1, 5, 4],
                        [1, 2, 5],
                        [2, 6, 5],
                        [2, 3, 6],
                        [3, 7, 6],
                        [3, 0, 4],
                        [3, 4, 7],
                        [1, 0, 3],
                        [2, 1, 3],
                        [7, 4, 5],
                        [5, 6, 7],
                    ]
                ),
            )
        )
        cell_data = {"medit:ref": np.array([[1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6]])}

        mesh = lnmmeshio.Mesh(points, cells, cell_data=cell_data)

        lnmmeshio.write_mesh(
            os.path.join(script_dir, "tmp", "dummy.mesh"),
            mesh,
            file_format="gmsh",
        )
        lnmmeshio.write_mesh(
            os.path.join(script_dir, "tmp", "dummy.stl"),
            mesh,
            file_format="mimicsstl",
        )

        # self.assertTrue(
        #    filecmp.cmp(
        #        os.path.join(script_dir, "data", "cube.mesh"),
        #        os.path.join(script_dir, "tmp", "dummy.mesh"),
        #    )
        # )
        self.assertTrue(
            filecmp.cmp(
                os.path.join(script_dir, "data", "cube.stl"),
                os.path.join(script_dir, "tmp", "dummy.stl"),
            )
        )

    def tearDown(self):
        # delete tmp folder
        # shutil.rmtree(os.path.join(script_dir, "tmp"))
        pass



from typing import List

import numpy as np

from ..node import Node
from .element import ElementQuad
from .line3 import Line3


class Quad8(ElementQuad):
    """
    Implementation of a quad8 element
    """

    ShapeName: str = "QUAD8"

    def __init__(self, el_type: str, nodes: List[Node]):
        """
        Base constructor of a Quad8 element
        """
        super(Quad8, self).__init__(el_type, Quad8.ShapeName, nodes)

        if len(nodes) != self.get_num_nodes():
            raise RuntimeError(
                "You tried to created a QUAD8 element with {0} nodes".format(len(nodes))
            )

    @classmethod
    def get_num_nodes(cls) -> int:
        """
        Get number of nodes of a QUAD8 element

        Returns:
            Number of nodes of a QUAD8 element = 8
        """
        return 8

    def get_faces(self) -> List["Quad8"]:
        """
        The face of a quad8 element is itself

        Returns:
            List of faces
        """
        return [self]

    def get_edges(self) -> List[Line3]:
        """
        Returns the list of all edges

        Returns:
            List of edges
        """
        return [
            Line3(None, [self.nodes[0], self.nodes[1], self.nodes[4]]),
            Line3(None, [self.nodes[1], self.nodes[2], self.nodes[5]]),
            Line3(None, [self.nodes[2], self.nodes[3], self.nodes[6]]),
            Line3(None, [self.nodes[3], self.nodes[0], self.nodes[7]]),
        ]

    @staticmethod
    def shape_fcns(xi):
        """
        Returns the value of the shape functions at the local coordinate xi
        """

        r = xi[0]
        s = xi[1]

        rp = 1.0 + r
        rm = 1.0 - r
        sp = 1.0 + s
        sm = 1.0 - s
        r2 = 1.0 - r * r
        s2 = 1.0 - s * s

        return np.array(
            [
                0.25 * (rm * sm - (r2 * sm + s2 * rm)),
                0.25 * (rp * sm - (r2 * sm + s2 * rp)),
                0.25 * (rp * sp - (s2 * rp + r2 * sp)),
                0.25 * (rm * sp - (r2 * sp + s2 * rm)),
                0.5 * r2 * sm,
                0.5 * s2 * rp,
                0.5 * r2 * sp,
                0.5 * s2 * rm,
            ]
        )



import os
import shutil
import unittest

import lnmmeshio

script_dir = os.path.dirname(os.path.realpath(__file__))


class TestOverride(unittest.TestCase):
    def setUp(self):
        if not os.path.isdir(os.path.join(script_dir, "tmp")):
            os.makedirs(os.path.join(script_dir, "tmp"))

        if not os.path.isfile(os.path.join(script_dir, "tmp", "cube.mesh")):
            shutil.copy(
                os.path.join(script_dir, "data", "cube.mesh"),
                os.path.join(script_dir, "tmp", "cube.mesh"),
            )

    def test_override_write(self):
        self.setUp()

        d = lnmmeshio.read(os.path.join(script_dir, "data", "dummy.dat"))

        with self.assertRaises(FileExistsError):
            lnmmeshio.write(
                os.path.join(script_dir, "tmp", "cube.mesh"), d, override=False
            )

        lnmmeshio.write(os.path.join(script_dir, "tmp", "cube.mesh"), d, override=True)

    def test_override_write_mesh(self):
        self.setUp()
        d = lnmmeshio.read_mesh(os.path.join(script_dir, "data", "dummy.mesh"))

        with self.assertRaises(FileExistsError):
            lnmmeshio.write_mesh(
                os.path.join(script_dir, "tmp", "cube.mesh"), d, override=False
            )

        lnmmeshio.write_mesh(
            os.path.join(script_dir, "tmp", "cube.mesh"), d, override=True
        )
